package de.tudresden.inf.st.mquat.solving.genetic;

import de.tudresden.inf.st.mquat.jastadd.model.Root;
import de.tudresden.inf.st.mquat.jastadd.model.Solution;
import de.tudresden.inf.st.mquat.solving.BenchmarkableSolver;
import de.tudresden.inf.st.mquat.solving.Solver;
import de.tudresden.inf.st.mquat.solving.SolvingException;
import de.tudresden.inf.st.mquat.solving.genetic.opt4j.Opt4jModule;
import de.tudresden.inf.st.mquat.solving.genetic.opt4j.custom.operator.copy.TreeCopyOperatorModule;
import de.tudresden.inf.st.mquat.solving.genetic.opt4j.custom.operator.crossover.TreeCrossoverOperatorModule;
import de.tudresden.inf.st.mquat.solving.genetic.opt4j.custom.operator.mutate.TreeMutateOperatorModule;
import de.tudresden.inf.st.mquat.solving.genetic.opt4j.custom.optimizer.TimeOutOptimizerModule;
import de.tudresden.inf.st.mquat.utils.StopWatch;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.opt4j.core.Individual;
import org.opt4j.core.optimizer.Archive;
import org.opt4j.core.start.Opt4JTask;
import org.opt4j.optimizers.ea.EvolutionaryAlgorithmModule;
import org.opt4j.optimizers.ea.Nsga2Module;
import org.opt4j.optimizers.ea.Spea2Module;

import java.util.concurrent.TimeUnit;

//import org.opt4j.viewer.ViewerModule;

public class GeneticSolver implements BenchmarkableSolver {

  private static final Logger logger = LogManager.getLogger(GeneticSolver.class);

  private Solution lastSolution;
  private long lastSolvingTime;

  private int solutionCounter;

  private StopWatch stopWatch;

  private long maxSolvingTime;
  private int generations;
  private int populationSize;
  private SelectorType selectorType;
  private boolean timedOut;

  public GeneticSolver() {
    this(SelectorType.NSGA2, 100, 100);
  }

  public GeneticSolver(SelectorType evaluatorType, int generations, int populationSize) {
    this.selectorType = evaluatorType;
    this.generations = generations;
    this.populationSize = populationSize;
    reset();

  }

  @Override
  public Solution solve(Root model) throws SolvingException {

    reset();

    stopWatch = StopWatch.start();

    if (model.getNumRequest() == 0) {
      return Solution.emptySolutionOf(model);
    }


    Opt4jModule.setModel(model);

    EvolutionaryAlgorithmModule ea = new EvolutionaryAlgorithmModule();
    ea.setGenerations(this.generations);

    // set population size
    ea.setAlpha(this.populationSize);

    Opt4jModule mquatModule = new Opt4jModule();

    mquatModule.setMaxSolvingTime(this.maxSolvingTime);

    TreeCrossoverOperatorModule crossover = new TreeCrossoverOperatorModule();
    TreeMutateOperatorModule mutate = new TreeMutateOperatorModule();
    TreeCopyOperatorModule copy = new TreeCopyOperatorModule();
    TimeOutOptimizerModule optimizerModule = new TimeOutOptimizerModule();

//    ViewerModule viewer = new ViewerModule();
//    viewer.setCloseOnStop(false);

    Opt4JTask task = new Opt4JTask(false);
    switch (this.selectorType) {
      case NSGA2:
        Nsga2Module nsga2 = new Nsga2Module();
        task.init(optimizerModule, ea, nsga2, mquatModule, crossover, mutate, copy);
//        task.init(optimizerModule, ea, nsga2, mquatModule, crossover, mutate, copy, viewer);
        break;
      case SPEA2:
        Spea2Module spea2 = new Spea2Module();
        task.init(optimizerModule, ea, spea2, mquatModule, crossover, mutate, copy);
//        task.init(optimizerModule, ea, spea2, mquatModule, crossover, mutate, copy, viewer);
    }
    try {
      task.execute();
      Archive archive = task.getInstance(Archive.class);
      for (Individual individual : archive) {
        // obtain the phenotype and objective, etc. of each individual
        Solution solution = (Solution) individual.getPhenotype();
        int validity = solution.evaluateValidity();
        double objective = solution.computeObjective();
        // TODO only works with minimized objective
        if (lastSolution == null || validity < lastSolution.evaluateValidity() || (validity == lastSolution.evaluateValidity() && objective < lastSolution.computeObjective())) {
          lastSolution = solution.deepCopy();
          logger.info("found a better solution with {} errors and an objective of {}.", validity, objective);
        }
      }
    } catch (Exception e) {
      e.printStackTrace();
    } finally {
      task.close();
    }

    if (lastSolution == null) {
      lastSolution = Solution.emptySolutionOf(model);
      logger.warn("Found no solution!");
    }

    lastSolvingTime = stopWatch.time(TimeUnit.MILLISECONDS);

    return lastSolution;
  }

  private void reset() {
    this.lastSolution = null;
    this.solutionCounter = 0;
    this.lastSolvingTime = 0;
    this.timedOut = false;
  }

  @Override
  public String getName() {
    return "genetic_" + this.selectorType.name();
  }

  @Override
  public long getLastSolvingTime() {
    return lastSolvingTime;
  }

  @Override
  public double getLastObjective() {
    if (lastSolution != null) {
      return lastSolution.computeObjective();
    } else {
      // TODO throw exception or do something reasonable
      return 0d;
    }
  }

  @Override
  public Solver setTimeout(long timeoutValue, TimeUnit timeoutUnit) {
    this.maxSolvingTime = timeoutUnit.toMillis(timeoutValue);
    return this;
  }

  @Override
  public boolean hadTimeout() {
    return this.timedOut;
  }

  public enum SelectorType {
    NSGA2,
    SPEA2
  }
}
